@@ -1,4 +1,5 @@
-/* $Id: compare.h,v 1.13 2002/02/08 22:27:46 bw Exp $ */
+/* Copyright (C) 2000-2004 Boris Wesslowski */
+/* $Id: compare.h,v 1.15 2004/03/21 09:42:55 bw Exp $ */
 
 #ifndef _COMPARE_H
 #define _COMPARE_H
@@ -9,6 +10,6 @@ struct conn_data *fwlw_mergesort(struct conn_data *list1);
 void sort_data(void);
 void build_list(void);
 int list_stats(void);
-void show_list(void);
+void show_list(FILE *fd);
 
 #endif
