@@ -1,4 +1,5 @@
-/* $Id: modes.c,v 1.54 2002/03/17 13:58:04 bw Exp $ */
+/* Copyright (C) 2000-2004 Boris Wesslowski */
+/* $Id: modes.c,v 1.79 2004/03/28 18:53:02 bw Exp $ */
 
 #include <stdio.h>
 #include <stdlib.h>
@@ -12,7 +13,11 @@
 #include <fcntl.h>
 #include <syslog.h>
 #include <signal.h>
+
+#ifdef HAVE_ZLIB
 #include <zlib.h>
+#endif
+
 #include "main.h"
 #include "parser.h"
 #include "output.h"
@@ -26,86 +31,133 @@
 
 extern struct options opt;
 extern struct conn_data *first;
+extern struct input_file *first_file;
 
-void mode_summary()
+void common_input_loop(int *linenum, int *hitnum, int *errnum, int *oldnum, int *exnum)
 {
-  char buf[BUFSIZE], nows[TIMESIZE], first_entry[TIMESIZE], last_entry[TIMESIZE];
-  FILE *output = NULL;
-  int retval, linenum = 0, hitnum = 0, hit = 0, errnum = 0, oldnum = 0, exnum = 0;
-  time_t now;
-  struct passwd *gen_user;
+  char buf[BUFSIZE];
+  int retval, hit = 0;
 
-  if (opt.std_in) {
-    if (opt.verbose)
-      fprintf(stderr, _("Using stdin as input\n"));
-
-    opt.inputfd = stdin;
-  } else {
-    if (opt.verbose)
-      fprintf(stderr, _("Opening input file '%s'\n"), opt.inputfile);
-
-    opt.inputfd = gzopen(opt.inputfile, "rb");
-    if (opt.inputfd == NULL) {
-      fprintf(stderr, "gzopen %s: %s\n", opt.inputfile, strerror(errno));
-      exit(EXIT_FAILURE);
-    }
-  }
-
-  if (opt.verbose)
-    fprintf(stderr, _("Processing\n"));
-
-  opt.line = xmalloc(sizeof(struct log_line));
-
-  if (opt.std_in) {
+#ifdef HAVE_ZLIB
+  if ((opt.std_in) || (opt.mode == REALTIME_RESPONSE)) {
+#endif
     retval = (fgets(buf, BUFSIZE, opt.inputfd) != NULL);
+#ifdef HAVE_ZLIB
   } else {
     retval = (gzgets(opt.inputfd, buf, BUFSIZE) != Z_NULL);
   }
+#endif
 
   while (retval) {
-    ++linenum;
+    *linenum += 1;
     hit = PARSE_NO_HIT;
-    hit = parse_line(buf, linenum);
+    hit = parse_line(buf, *linenum);
     opt.repeated = 0;
     switch (hit) {
     case PARSE_OK:
-      ++hitnum;
+      *hitnum += 1;
       opt.repeated = 1;
       break;
     case PARSE_WRONG_FORMAT:
-      ++errnum;
+      *errnum += 1;
       break;
     case PARSE_TOO_OLD:
-      ++oldnum;
+      *oldnum += 1;
       break;
     case PARSE_EXCLUDED:
-      ++hitnum;
-      ++exnum;
-      break;
+      *hitnum += 1;
+      *exnum += 1;
     }
 
-    if (opt.std_in) {
+#ifdef HAVE_ZLIB
+    if ((opt.std_in) || (opt.mode == REALTIME_RESPONSE)) {
+#endif
       retval = (fgets(buf, BUFSIZE, opt.inputfd) != NULL);
+#ifdef HAVE_ZLIB
     } else {
       retval = (gzgets(opt.inputfd, buf, BUFSIZE) != Z_NULL);
     }
+#endif
   }
+}
 
-  if (opt.verbose == 2)
-    fprintf(stderr, "\n");
+void mode_summary()
+{
+  char nows[TIMESIZE], first_entry[TIMESIZE], last_entry[TIMESIZE], *input = NULL, last_file = 0;
+  FILE *output = NULL;
+  int retval, linenum = 0, hitnum = 0, errnum = 0, oldnum = 0, exnum = 0;
+  time_t now;
+  struct passwd *gen_user;
+  struct input_file *file;
 
-  if(!opt.std_in) {
-    if (opt.verbose)
-      fprintf(stderr, _("Closing '%s'\n"), opt.inputfile);
+  opt.line = xmalloc(sizeof(struct log_line));
 
-    retval = gzclose(opt.inputfd);
-    if (retval != 0) {
-      if (retval != Z_ERRNO) {
-	fprintf(stderr, "gzclose %s: %s\n", opt.inputfile, gzerror(opt.inputfd, &retval));
+  file = first_file;
+  while (last_file == 0) {
+    if (opt.std_in) {
+      if (opt.verbose)
+	fprintf(stderr, _("Using stdin as input\n"));
+
+      opt.inputfd = stdin;
+    } else {
+      input = file->name;
+      if (opt.verbose)
+	fprintf(stderr, _("Opening input file '%s'\n"), input);
+
+#ifdef HAVE_ZLIB
+      opt.inputfd = gzopen(input, "rb");
+#else
+      opt.inputfd = fopen(input, "r");
+#endif
+      if (opt.inputfd == NULL) {
+#ifdef HAVE_ZLIB
+	fprintf(stderr, "gzopen %s: %s\n", input, strerror(errno));
+#else
+	fprintf(stderr, "fopen %s: %s\n", input, strerror(errno));
+#endif
+	exit(EXIT_FAILURE);
+      }
+    }
+
+    if (opt.verbose)
+      fprintf(stderr, _("Processing\n"));
+
+    common_input_loop(&linenum, &hitnum, &errnum, &oldnum, &exnum);
+
+    if (opt.verbose == 2)
+      fprintf(stderr, "\n");
+    if (errnum && opt.verbose) {
+      fprintf(stderr, _("Unrecognized entries or tokens can be submitted at\n"));
+      fprintf(stderr, "http://fwlogwatch.inside-security.de/unrecognized.php\n");
+    }
+
+    if(opt.std_in) {
+      last_file++;
+    } else {
+      if (opt.verbose)
+	fprintf(stderr, _("Closing '%s'\n"), input);
+
+#ifndef HAVE_ZLIB
+      retval = fclose(opt.inputfd);
+      if (retval == EOF) {
+	perror("fclose");
+#else
+      retval = gzclose(opt.inputfd);
+      if (retval != 0) {
+	if (retval != Z_ERRNO) {
+	  fprintf(stderr, "gzclose %s: %s\n", input, gzerror(opt.inputfd, &retval));
+	} else {
+	  perror("gzclose");
+	}
+#endif
+	exit(EXIT_FAILURE);
+      }
+
+      if (file->next != NULL) {
+	file = file->next;
       } else {
-	perror("gzclose");
+	last_file++;
       }
-      exit(EXIT_FAILURE);
     }
   }
 
@@ -115,17 +167,18 @@ void mode_summary()
     fprintf(stderr, _("Sorting data\n"));
 
   if(first != NULL) {
-    opt.sortfield = SORT_START_TIME;
+    opt.sortfield = SORT_END_TIME;
     opt.sortmode = ORDER_DESCENDING;
     first = fwlw_mergesort(first);
     if(opt.verbose == 2)
       fprintf(stderr, ".");
-    strftime(last_entry, TIMESIZE, "%b %d %H:%M:%S", localtime(&first->start_time));
+    strftime(last_entry, TIMESIZE, _("%b %d %H:%M:%S"), localtime(&first->end_time));
+    opt.sortfield = SORT_START_TIME;
     opt.sortmode = ORDER_ASCENDING;
     first = fwlw_mergesort(first);
     if(opt.verbose == 2)
       fprintf(stderr, ".");
-    strftime(first_entry, TIMESIZE, "%b %d %H:%M:%S", localtime(&first->start_time));
+    strftime(first_entry, TIMESIZE, _("%b %d %H:%M:%S"), localtime(&first->start_time));
   } else {
     first_entry[0] = '\0';
   }
@@ -139,121 +192,151 @@ void mode_summary()
     if (opt.verbose)
       fprintf(stderr, _("Opening output file '%s'\n"), opt.outputfile);
 
-    fflush(stdout);
     output = freopen(opt.outputfile, "w", stdout);
     if (output == NULL) {
       fprintf(stderr, "freopen %s: %s\n", opt.outputfile, strerror(errno));
       exit(EXIT_FAILURE);
     }
+  } else if ((opt.mode != INTERACTIVE_REPORT) && (opt.recipient[0] != '\0')) {
+    char buf[BUFSIZE];
+
+    if(opt.verbose)
+      fprintf(stderr, _("Sending\n"));
+
+    snprintf(buf, BUFSIZE, "%s -t", P_SENDMAIL);
+    output = popen(buf, "w");
+    if(output == NULL) {
+      perror("popen");
+      exit(EXIT_FAILURE);
+    }
+
+    generate_email_header(output);
+    fflush(output);
+  } else {
+    output = stdout;
   }
 
   if (opt.html) {
-    output_html_header();
+    output_html_header(fileno(output));
+    fprintf(output, "<p>\n");
   } else {
-    printf("%s\n", opt.title);
+    fprintf(output, "%s\n", opt.title);
   }
 
   now = time(NULL);
-  strftime(nows, TIMESIZE, "%a %b %d %H:%M:%S %Z %Y", localtime(&now));
-  printf(_("Generated %s by "), nows);
+  strftime(nows, TIMESIZE, _("%A %B %d %H:%M:%S %Z %Y"), localtime(&now));
+  fprintf(output, _("Generated %s by "), nows);
 
   gen_user = getpwuid(getuid());
   if (gen_user != NULL) {
     if (gen_user->pw_gecos[0] != '\0') {
-      printf("%s.\n", gen_user->pw_gecos);
+      fprintf(output, "%s.\n", gen_user->pw_gecos);
     } else {
-      printf("%s.\n", gen_user->pw_name);
+      fprintf(output, "%s.\n", gen_user->pw_name);
     }
   } else {
-    printf(_("an unknown user.\n"));
+    fprintf(output, _("an unknown user.\n"));
   }
 
   if (opt.html)
-    printf("<br>\n");
+    fprintf(output, "<br />\n");
 
-  printf("%d ", hitnum);
+  fprintf(output, "%d ", hitnum);
   if (oldnum > 0) {
-    printf(_("(and %d older than %d seconds) "), oldnum, opt.recent);
+    fprintf(output, _("(and %d older than %d seconds) "), oldnum, opt.recent);
   }
   if (errnum > 0) {
-    printf(_("(and %d malformed) "), errnum);
+    fprintf(output, _("(and %d malformed) "), errnum);
+  }
+  if (opt.filecount == 1) {
+    fprintf(output, _("of %d entries in the file \"%s\" are packet logs, "), linenum, input);
+  } else if (opt.filecount == 0) {
+    fprintf(output, _("of %d entries in standard input are packet logs, "), linenum);
+  } else {
+    fprintf(output, _("of %d entries in %d input files are packet logs, "), linenum, opt.filecount);
   }
-  printf(_("of %d entries in the file "), linenum);
-  printf(_("\"%s\" are packet logs, "), opt.inputfile);
   retval = list_stats();
   if(retval == 1) {
-    printf(_("one has unique characteristics.\n"));
+    fprintf(output, _("one has unique characteristics.\n"));
   } else {
-    printf(_("%d have unique characteristics.\n"), retval);
+    fprintf(output, _("%d have unique characteristics.\n"), retval);
   }
 
   if (exnum != 0) {
     if (opt.html)
-      printf("<br>\n");
+      fprintf(output, "<br />\n");
 
     if(exnum == 1) {
-      printf(_("One entry was excluded by configuration.\n"));
+      fprintf(output, _("One entry was excluded by configuration.\n"));
     } else {
-      printf(_("%d entries were excluded by configuration.\n"), exnum);
+      fprintf(output, _("%d entries were excluded by configuration.\n"), exnum);
     }
   }
 
   if (opt.html)
-    printf("<br>\n");
+    fprintf(output, "<br />\n");
 
   if (first_entry[0] != '\0') {
-    printf(_("First packet log entry: %s, last: %s.\n"), first_entry, last_entry);
+    fprintf(output, _("First packet log entry: %s, last: %s.\n"), first_entry, last_entry);
   } else {
-    printf(_("No valid time entries found.\n"));
+    fprintf(output, _("No valid time entries found.\n"));
   }
 
   if(!opt.loghost) {
     if(opt.html)
-      printf("<br>\n");
+      fprintf(output, "<br />\n");
 
-    printf(_("All entries were logged by the same host: \"%s\".\n"), opt.hostname);
+    fprintf(output, _("All entries were logged by the same host: \"%s\".\n"), opt.hostname);
   }
 
   if(!opt.chains) {
     if(opt.html)
-      printf("<br>\n");
+      fprintf(output, "<br />\n");
 
-    printf(_("All entries are from the same chain: \"%s\".\n"), opt.chainlabel);
+    fprintf(output, _("All entries are from the same chain: \"%s\".\n"), opt.chainlabel);
   }
 
   if(!opt.branches) {
     if(opt.html)
-      printf("<br>\n");
+      fprintf(output, "<br />\n");
 
-    printf(_("All entries have the same target: \"%s\".\n"), opt.branchname);
+    fprintf(output, _("All entries have the same target: \"%s\".\n"), opt.branchname);
   }
 
   if(!opt.ifs) {
     if(opt.html)
-      printf("<br>\n");
+      fprintf(output, "<br />\n");
 
-    printf(_("All entries are from the same interface: \"%s\".\n"), opt.interface);
+    fprintf(output, _("All entries are from the same interface: \"%s\".\n"), opt.interface);
   }
 
   if(opt.least > 1) {
     if(opt.html)
-      printf("<br>\n");
+      fprintf(output, "<br />\n");
 
-    printf(_("Only entries with a count of at least %d are shown.\n"), opt.least);
+    fprintf(output, _("Only entries with a count of at least %d are shown.\n"), opt.least);
+  }
+
+  if(opt.max) {
+    if(opt.html)
+      fprintf(output, "<br />\n");
+
+    fprintf(output, _("Only the top %d entries are shown.\n"), opt.max);
   }
 
   if (opt.html)
-    output_html_table();
+    output_html_table(output);
   else
-    printf("\n");
+    fprintf(output, "\n");
 
   if(opt.mode == INTERACTIVE_REPORT)
-    printf(_("Reporting threshold: %d\n\n"), opt.threshold);
+    fprintf(output, _("Reporting threshold: %d\n\n"), opt.threshold);
 
   if(opt.whois_lookup)
     whois_connect(RADB);
 
-  show_list();
+  show_list(output);
+  fflush(output);
 
   if(opt.whois_lookup)
     whois_close();
@@ -261,12 +344,17 @@ void mode_summary()
   if(opt.mode == INTERACTIVE_REPORT)
     report();
 
-  if (opt.html)
-    output_html_footer();
+  if (opt.html) {
+    fprintf(output, "</table>\n");
+    fflush(output);
+    output_html_footer(fileno(output));
+  }
 
   free_conn_data();
   free_dns_cache();
+  free_whois();
   free_exclude_data();
+  free_input_file();
 
   if (opt.use_out) {
     if (opt.verbose)
@@ -275,7 +363,11 @@ void mode_summary()
     retval = fclose(output);
     if (retval == EOF) {
       perror("fclose");
-      exit(EXIT_FAILURE);
+    }
+  } else if ((opt.mode != INTERACTIVE_REPORT) && (opt.recipient[0] != '\0')) {
+    retval = pclose(output);
+    if (retval == -1) {
+      perror("pclose");
     }
   }
 }
@@ -312,9 +404,9 @@ void mode_rt_response_open()
   if(opt.std_in) {
     opt.inputfd = stdin;
   } else {
-    opt.inputfd = fopen(opt.inputfile, "r");
+    opt.inputfd = fopen(first_file->name, "r");
     if (opt.inputfd == NULL) {
-      syslog(LOG_NOTICE, "fopen %s: %s", opt.inputfile, strerror(errno));
+      syslog(LOG_NOTICE, "fopen %s: %s", first_file->name, strerror(errno));
       log_exit(EXIT_FAILURE);
     }
   }
@@ -327,11 +419,11 @@ void mode_rt_response_reopen_log()
   if(opt.std_in) {
     syslog(LOG_NOTICE, _("SIGUSR1 caught, reading input from stdin, no need to reopen log file"));
   } else {
-    syslog(LOG_NOTICE, _("SIGUSR1 caught, reopening log file %s"), opt.inputfile);
+    syslog(LOG_NOTICE, _("SIGUSR1 caught, reopening log file %s"), first_file->name);
 
     retval = fclose(opt.inputfd);
     if(retval == EOF)
-      syslog(LOG_NOTICE, "fclose %s: %s", opt.inputfile, strerror(errno));
+      syslog(LOG_NOTICE, "fclose %s: %s", first_file->name, strerror(errno));
 
     mode_rt_response_open();
     signal(SIGUSR1, mode_rt_response_reopen_log);
@@ -340,22 +432,23 @@ void mode_rt_response_reopen_log()
 
 void mode_rt_response_core()
 {
-  char buf[BUFSIZE];
-  int retval;
+  int retval, linenum = 0, hitnum = 0, ignored = 0;
   struct stat info;
   off_t size = 0;
   fd_set rfds;
   struct timeval tv;
 
-  if(!opt.std_in) {
+  if((!opt.std_in) && (!opt.stateful_start)) {
     retval = fstat(fileno(opt.inputfd), &info);
     if (retval == -1) {
-      syslog(LOG_NOTICE, "fstat %s: %s", opt.inputfile, strerror(errno));
+      syslog(LOG_NOTICE, "fstat %s: %s", first_file->name, strerror(errno));
       log_exit(EXIT_FAILURE);
     }
     size = info.st_size;
   }
 
+  opt.line = xmalloc(sizeof(struct log_line));
+
   while (1) {
     if(opt.status) {
       FD_ZERO(&rfds);
@@ -370,34 +463,26 @@ void mode_rt_response_core()
 	}
       }
       if (retval > 0) {
-	handshake();
+	handshake(linenum, hitnum, ignored);
       }
     } else {
       sleep(1);
     }
 
-    remove_old();
+    remove_old(RESP_REMOVE_OPC|RESP_REMOVE_OHS);
     if(opt.std_in) {
-      while (fgets(buf, BUFSIZE, opt.inputfd)) {
-	opt.line = xmalloc(sizeof(struct log_line));
-	parse_line(buf, 0);
-	free(opt.line);
-      }
+      common_input_loop(&linenum, &hitnum, &ignored, &ignored, &ignored);
       look_for_alert();
     } else {
       retval = fstat(fileno(opt.inputfd), &info);
       if (retval == -1) {
-	syslog(LOG_NOTICE, "fstat %s: %s", opt.inputfile, strerror(errno));
+	syslog(LOG_NOTICE, "fstat %s: %s", first_file->name, strerror(errno));
 	log_exit(EXIT_FAILURE);
       }
       if(size != info.st_size) {
 	size = info.st_size;
 	clearerr(opt.inputfd);
-	while (fgets(buf, BUFSIZE, opt.inputfd)) {
-	  opt.line = xmalloc(sizeof(struct log_line));
-	  parse_line(buf, 0);
-	  free(opt.line);
-	}
+	common_input_loop(&linenum, &hitnum, &ignored, &ignored, &ignored);
 	look_for_alert();
       }
     }
@@ -528,7 +613,7 @@ void mode_rt_response()
 
     pwe = getpwnam(opt.run_as);
     if(pwe == NULL) {
-      syslog(LOG_NOTICE, "User to run as was not found");
+      syslog(LOG_NOTICE, _("User to run as was not found"));
       log_exit(EXIT_FAILURE);
     }
     olduid = getuid();
@@ -543,9 +628,9 @@ void mode_rt_response()
       syslog(LOG_NOTICE, "setuid: %s", strerror(errno));
       log_exit(EXIT_FAILURE);
     }
-    syslog(LOG_NOTICE, "Changed uid from %d to %d, gid from %d to %d", olduid, getuid(), oldgid, getgid());
+    syslog(LOG_NOTICE, _("Changed uid from %d to %d, gid from %d to %d"), olduid, getuid(), oldgid, getgid());
   } else {
-    syslog(LOG_NOTICE, "Running with uid %d, gid %d", getuid(), getgid());
+    syslog(LOG_NOTICE, _("Running with uid %d, gid %d"), getuid(), getgid());
   }
 
   if(opt.threshold == 1) {
@@ -562,14 +647,14 @@ void mode_rt_response()
 	   opt.recent/3600);
   }
 
-  syslog(LOG_NOTICE, _("Response mode: log%s%s"),
+  syslog(LOG_NOTICE, _("Response mode: Log%s%s"),
 	 (opt.response & OPT_NOTIFY)?_(", notify"):"",
 	 (opt.response & OPT_RESPOND)?_(", respond"):"");
 
-  if(!opt.std_in) {
+  if((!opt.std_in) && (!opt.stateful_start)) {
     retval = fseek(opt.inputfd, 0, SEEK_END);
     if (retval == -1) {
-      syslog(LOG_NOTICE, "fseek %s: %s", opt.inputfile, strerror(errno));
+      syslog(LOG_NOTICE, "fseek %s: %s", first_file->name, strerror(errno));
       log_exit(EXIT_FAILURE);
     }
   }
@@ -581,80 +666,116 @@ void mode_rt_response()
 
 void mode_show_log_times()
 {
-  char first_entry[TIMESIZE], last_entry[TIMESIZE], buf[BUFSIZE], month[3];
-  int retval = 0, retval2, day, hour, minute, second, linenum = 0;
-  FILE *input;
+  char buf[BUFSIZE], stime[TIMESIZE], month[3], *input = NULL, last_file = 0;
+  int retval = 0, loop, day, hour, minute, second, linenum = 0;
+  struct input_file *file;
+  time_t first = 0, last = 0;
 
-  if (opt.std_in) {
-    opt.inputfd = stdin;
-    input = 0;
-
-    if (opt.verbose)
-      fprintf(stderr, _("Reading standard input\n"));
-  } else {
-    input = gzopen(opt.inputfile, "rb");
-    if (input == NULL) {
-      fprintf(stderr, "gzopen %s: %s\n", opt.inputfile, strerror(errno));
-      exit(EXIT_FAILURE);
-    }
-
-    if (opt.verbose)
-      fprintf(stderr, _("Reading '%s'\n"), opt.inputfile);
-  }
-
-  if (opt.std_in) {
-    retval2 = (fgets(buf, BUFSIZE, opt.inputfd) != NULL);
-  } else {
-    retval2 = (gzgets(input, buf, BUFSIZE) != Z_NULL);
-  }
-
-  while ((retval != 5) && (retval2)) {
-    retval = sscanf(buf, "%3s %2d %2d:%2d:%2d ", month, &day, &hour, &minute, &second);
-    linenum++;
+  opt.line = xmalloc(sizeof(struct log_line));
 
+  file = first_file;
+  while (last_file == 0) {
     if (opt.std_in) {
-      retval2 = (fgets(buf, BUFSIZE, opt.inputfd) != NULL);
+      opt.inputfd = stdin;
+
+      if (opt.verbose)
+	fprintf(stderr, _("Reading standard input\n"));
     } else {
-      retval2 = (gzgets(input, buf, BUFSIZE) != Z_NULL);
+      input = file->name;
+#ifdef HAVE_ZLIB
+      opt.inputfd = gzopen(input, "rb");
+#else
+      opt.inputfd = fopen(input, "r");
+#endif
+      if (opt.inputfd == NULL) {
+#ifdef HAVE_ZLIB
+	fprintf(stderr, "gzopen %s: %s\n", input, strerror(errno));
+#else
+	fprintf(stderr, "fopen %s: %s\n", input, strerror(errno));
+#endif
+	exit(EXIT_FAILURE);
+      }
+
+      if (opt.verbose)
+	fprintf(stderr, _("Reading '%s'\n"), input);
     }
-  }
 
-  if(retval == 5) {
-    snprintf(first_entry, TIMESIZE, "%s %02d %02d:%02d:%02d", month, day, hour, minute, second);
+#ifdef HAVE_ZLIB
+    if (opt.std_in) {
+#endif
+      loop = (fgets(buf, BUFSIZE, opt.inputfd) != NULL);
+#ifdef HAVE_ZLIB
+    } else {
+      loop = (gzgets(opt.inputfd, buf, BUFSIZE) != Z_NULL);
+    }
+#endif
 
-    retval = 0;
-    while (retval2) {
-      retval = sscanf(buf, "%3s %2d %2d:%2d:%2d ", month, &day, &hour, &minute, &second);
+    while (loop) {
       linenum++;
+      retval = sscanf(buf, "%3s %2d %2d:%2d:%2d ", month, &day, &hour, &minute, &second);
+      if(retval == 5) {
+	build_time(month, day, hour, minute, second);
+	if (first == 0)
+	  first = last = opt.line->time;
+	if (opt.line->time < first)
+	  first = opt.line->time;
+	if (opt.line->time > last)
+	  last = opt.line->time;
+      }
+
+#ifdef HAVE_ZLIB
       if (opt.std_in) {
-	retval2 = (fgets(buf, BUFSIZE, opt.inputfd) != NULL);
+#endif
+	loop = (fgets(buf, BUFSIZE, opt.inputfd) != NULL);
+#ifdef HAVE_ZLIB
       } else {
-	retval2 = (gzgets(input, buf, BUFSIZE) != Z_NULL);
+	loop = (gzgets(opt.inputfd, buf, BUFSIZE) != Z_NULL);
       }
+#endif
     }
 
-    printf(_("# of lines : %d\n"), linenum);
-    printf(_("First entry: %s\n"), first_entry);
-    if (retval == 5) {
-      snprintf(last_entry, TIMESIZE, "%s %02d %02d:%02d:%02d", month, day, hour, minute, second);
-      printf(_("Last entry : %s\n"), last_entry);
+    if(opt.std_in) {
+      last_file++;
+    } else {
+      if (opt.verbose)
+	fprintf(stderr, _("Closing '%s'\n"), input);
+
+#ifndef HAVE_ZLIB
+      retval = fclose(opt.inputfd);
+      if (retval == EOF) {
+	perror("fclose");
+#else
+      retval = gzclose(opt.inputfd);
+      if (retval != 0) {
+	if (retval != Z_ERRNO) {
+	  fprintf(stderr, "gzclose %s: %s\n", input, gzerror(opt.inputfd, &retval));
+	} else {
+	  perror("gzclose");
+	}
+#endif
+	exit(EXIT_FAILURE);
+      }
+
+      if (file->next != NULL) {
+	file = file->next;
+      } else {
+	last_file++;
+      }
     }
-  } else {
+  }
+
+  printf(_("Number of files: %d\n"), opt.filecount);
+  printf(_("Number of lines: %d\n"), linenum);
+  if (first == 0) {
     printf(_("No valid time entries found.\n"));
+  } else {
+    strftime(stime, TIMESIZE, _("%b %d %H:%M:%S"), localtime(&first));
+    printf(_("First entry: %s\n"), stime);
+    strftime(stime, TIMESIZE, _("%b %d %H:%M:%S"), localtime(&last));
+    printf(_("Last entry : %s\n"), stime);
+    output_timediff(first, last, stime);
+    printf(_("Difference : %s\n"), stime);
   }
 
-  if(!opt.std_in) {
-    if (opt.verbose)
-      fprintf(stderr, _("Closing '%s'\n"), opt.inputfile);
-
-    retval = gzclose(input);
-    if (retval != 0) {
-      if (retval != Z_ERRNO) {
-	fprintf(stderr, "gzclose %s: %s\n", opt.inputfile, gzerror(input, &retval));
-      } else {
-	perror("gzclose");
-      }
-      exit(EXIT_FAILURE);
-    }
-  }
+  free(opt.line);
 }
